
import datetime
import traceback
import json
import numpy as np
import pandas as pd
from typing import List, Dict, Optional, Tuple, Union, Set

from ....constant import SectorType, IndexPriceSource, INDEX_PRICE_EXTRA_TRADE_DAYS, FUND_NAV_EXTRA_TRADE_DAYS
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.basic_models import *
from ...manager.fund_tag_update import fund_tag
# from ..raw.raw_data_helper import RawDataHelper
from .basic_data_helper import BasicDataHelper


class BasicDataPart1:

    _REPORT_DATE_LIST = ['0331', '0630', '0930', '1231']

    def __init__(self, data_helper: BasicDataHelper):
        self._data_helper = data_helper
        self._raw_data_api = RawDataApi()
        self._basic_data_api = BasicDataApi()
        self._index_info_df = self._basic_data_api.get_index_info()
        self._fund_info_df = self._basic_data_api.get_fund_info()

    def _status_mapping(self, x):
        if x == 'Open':
            return 0
        elif x == 'Suspended':
            return 1
        elif x == 'Limited':
            return 2
        elif x == 'Close':
            return 3
        else:
            return None

    def _find_tag(self, symbol, wind_class_II):
        if '沪深300' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股大盘'
        elif '中证500' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return 'A股中盘'
        elif '标普500' in symbol:
            return '美股大盘'
        elif '创业板' in symbol and wind_class_II in ['普通股票型基金', '增强指数型基金', '被动指数型基金']:
            return '创业板'
        elif '德国' in symbol:
            return '德国大盘'
        elif '日本' in symbol or '日经' in symbol:
            return '日本大盘'
        elif (('国债' in symbol or '利率' in symbol or '金融债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '利率债'
        elif (('信用' in symbol or '企债' in symbol or '企业债' in symbol)
              and wind_class_II in ['短期纯债型基金', '中长期纯债型基金', '被动指数型债券基金']):
            return '信用债'
        elif '黄金' in symbol:
            return '黄金'
        elif '原油' in symbol or '石油' in symbol or '油气' in symbol:
            return '原油'
        elif ('地产' in symbol or '金融' in symbol) and ('美国'not in symbol):
            return '房地产'
        else:
            return 'null'

    def fund_info(self):
        # Update manually
        # Not verified
        track_index_df = pd.read_csv('./data/fund_track_index.csv', index_col=0)
        wind_fund_info = self._raw_data_api.get_wind_fund_info()
        fund_fee = self._raw_data_api.get_fund_fee()
        wind_fund_info['order_book_id'] = [_.split('!')[0].split('.')[0] for _ in wind_fund_info['wind_id'] ]
        res = []
        for i in wind_fund_info['wind_id']:
            if not '!' in i:
                res.append(0)
            else:
                res_i = int(i.split('!')[1].split('.')[0])
                res.append(res_i)
        wind_fund_info['transition'] = res
        wind_fund_info['fund_id'] = [o + '!' + str(t) for o, t in zip(wind_fund_info['order_book_id'], wind_fund_info['transition'])]
        wind_fund_info = wind_fund_info.set_index('fund_id')
        fund_fee = fund_fee.drop(['id','desc_name'], axis = 1).set_index('fund_id')
        wind_fund_info = wind_fund_info.join(fund_fee)
        wind_fund_info['fund_id'] = wind_fund_info.index
        wind_fund_info['asset_type'] = [self._find_tag(symbol, wind_class_II) for symbol, wind_class_II in zip(wind_fund_info['desc_name'],wind_fund_info['wind_class_2'])]
        wind_fund_info = wind_fund_info[[i.split('.')[1] == 'OF' for i in wind_fund_info['wind_id']]]
        wind_fund_info = wind_fund_info.set_index('wind_id')
        dic = {k:v for k,v in zip(self._index_info_df['desc_name'], self._index_info_df['index_id'])}
        wind_fund_info = wind_fund_info.join(track_index_df[['track_index']])
        wind_fund_info['wind_id'] = wind_fund_info.index
        wind_fund_info['track_index'] = wind_fund_info['track_index'].map(lambda x: dic.get(x,'none'))
        self._data_helper._upload_basic(wind_fund_info, FundInfo.__table__.name)

    def fund_info_from_raw(self, fund_id_list: Set[str], benchmark_profile: pd.Series):
        def _invest_type_class_to_wind_class_1(x: pd.Series) -> str:
            if x.invest_type_class == '其他基金':
                if 'QDII' in x.desc_name:
                    return '国际(QDII)基金'
            return x.invest_type_class

        def _invest_type_class_to_wind_class_2(invest_type_class: str) -> str:
            '''Choice的基金二级类型转换为wind的'''

            if invest_type_class == '混合债券型基金（一级）':
                return '混合债券型一级基金'
            elif invest_type_class == '长期纯债型基金':
                return '中长期纯债型基金'
            elif invest_type_class == '中短期纯债型基金':
                return '短期纯债型基金'
            elif invest_type_class == '混合债券型基金（二级）':
                return '混合债券型二级基金'
            elif invest_type_class == '被动指数型基金':
                return '被动指数型债券基金'
            elif invest_type_class in ('QDII基金', '封闭式基金', '保本型基金'):
                # error case
                return
            else:
                return invest_type_class

        def _lambda_structure_type(x: pd.Series) -> int:
            if x.base_fund_id is None:
                return 0
            if x.base_fund_id == x.wind_id:
                return 1
            # 到这里应该必然是分级基金，所以直接判断A,B即可
            elif 'A' in x.desc_name:
                return 2
            elif 'B' in x.desc_name:
                return 3
            else:
                # 分级基金A,B份额可能名字里不带A,B，之后需要手工修改下
                return 4

        def _lambda_index_id_new(x: pd.Series, benchmark: pd.DataFrame) -> Optional[str]:
            try:
                return fund_tag(x.base_fund_id, x.desc_name, x.is_open, x.currency, x.wind_class_2, benchmark.at[x.wind_id, 'benchmark_s_raw'], x.fund_id)
            except KeyError:
                print(f'{x.fund_id} has no benchmark info')
                return

        def _lambda_fund_manager(x: str) -> pd.Series:
            manager_list: List[str] = x.split(',')
            return pd.Series(['\r\n'.join(manager_list), manager_list[0]])

        def _lambda_benchmark(x: str) -> pd.Series:
            def _parse_index(part: str):
                if '×' in part:
                    es = part.split('×')
                elif '*' in part:
                    es = part.split('*')
                elif '╳' in part:
                    es = part.split('╳')
                else:
                    return part

                if '%' in es[0]:
                    return es[1]
                else:
                    return es[0]

            if x is None:
                return
            benchmarks: List[str] = x.split('+')
            if len(benchmarks) == 1:
                second = None
            else:
                second = _parse_index(benchmarks[1])
            return pd.Series([_parse_index(benchmarks[0]), second])

        def _lambda_fee(x: str, em_fund_fee: pd.DataFrame) -> pd.Series:
            def _parse_fee(fee: str):
                if fee is None:
                    return
                cheapest_fee = fee.split(';')[0]
                es = cheapest_fee.split(':')
                if len(es) > 1:
                    value = es[1]
                else:
                    value = es[0]
                if value[-1] == '%':
                    value = float(value[:-1]) / 100
                else:
                    try:
                        value = float(value) / 100
                    except ValueError:
                        return
                return value

            fee = em_fund_fee.loc[x, ['manage_fee', 'trustee_fee', 'purchase_fee', 'redeem_fee']]
            return fee.transform(_parse_fee)

        # 获取raw中的基金基础信息
        df = self._raw_data_api.get_em_fund_info(fund_id_list).drop(columns='_update_time')
        # 其中一部分列可以直接使用，改一下名字即可
        df = df.rename(columns={'em_id': 'wind_id', 'name': 'desc_name', 'manager_id': 'pre_fund_manager'})
        # is open
        df['is_open'] = 1
        key_words = ['封闭','定期','定开']
        key_words = '|'.join(key_words)
        _wind_ids = df[df['desc_name'].str.contains(key_words)].wind_id
        df = df.set_index('wind_id')
        df.loc[_wind_ids,'is_open'] = 0
        df = df.reset_index()
        # 没有end_date的话给一个默认值
        df['end_date'] = df['end_date'].where(df.end_date.notna(), '2040-12-31')
        # 这里可能还需要加上港币
        df['currency'] = df.currency.map({'人民币元': 'CNY', '美元': 'USD'})
        # 基金经理
        df[['manager_id', 'fund_manager']] = df.pre_fund_manager.apply(_lambda_fund_manager)
        # 新基金都是!0，这里需要检查是否有相同order_book_id的旧基金
        df['order_book_id'] = df.wind_id.transform(lambda x: x.split('.')[0])
        df['transition'] = 0
        df['fund_id'] = df.order_book_id.transform(lambda x: x + '!0')
        # Choice的基金一二级分类转换为wind的
        df['wind_class_1'] = df[['desc_name', 'invest_type_class_2']].transform(_invest_type_class_to_wind_class_1, axis=1)
        df['wind_class_2'] = df.invest_type_class_2.transform(_invest_type_class_to_wind_class_2)
        df['asset_type'] = df.apply(lambda x: self._find_tag(x.desc_name, x.wind_class_2), axis=1)
        df['structure_type'] = df.apply(_lambda_structure_type, axis=1)
        df['is_etf'] = df.desc_name.transform(lambda x: 1 if ('ETF' in x and '联接' not in x) else 0)
        fund_benchmark = self._basic_data_api.get_fund_benchmark(fund_id_list).set_index('em_id')
        df['index_id_new'] = df.apply(_lambda_index_id_new, axis=1, benchmark=fund_benchmark)
        em_fund_fee = self._raw_data_api.get_em_fund_fee(fund_id_list).drop(columns='_update_time').set_index('em_id')
        df[['manage_fee', 'trustee_fee', 'purchase_fee', 'redeem_fee']] = df.wind_id.transform(_lambda_fee, em_fund_fee=em_fund_fee)
        df = df.set_index('wind_id').join(em_fund_fee[['subscr_fee', 'purchase_fee', 'redeem_fee', 'service_fee']].rename(columns={'subscr_fee': 'subscr_fee_detail', 'purchase_fee': 'purchase_fee_detail', 'redeem_fee': 'redeem_fee_detail'}))
        df['benchmark'] = benchmark_profile
        # 拆解上边的benchmark
        df[['benchmark_1', 'benchmark_2']] = df.benchmark.apply(_lambda_benchmark)
        df = df.drop(columns='invest_type_class_2').reset_index()
        # 将场内变为场外的
        df['wind_id'] = df.wind_id.transform(lambda x: x.split('.')[0] + '.OF' if '.OF' not in x else x)
        self._data_helper._upload_basic(df, FundInfo.__table__.name)

    def index_info(self):
        # Update manually
        # Not verified
        df = pd.read_csv('./data/index_info.csv')
        self._data_helper._upload_basic(df, IndexInfo.__table__.name)

    def index_info_from_raw(self, index_list: Union[Tuple[str], List[str]], index_ids: List[str]):
        assert len(index_list) == len(index_ids), 'index list length should match the ids'

        df = self._raw_data_api.get_em_index_info(index_list).drop(columns=['name', 'publish_date', '_update_time'])
        assert df.shape[0] == len(index_ids), 'index info length should match the ids'
        df = df.rename(columns={'short_name': 'desc_name'})
        # 先全体置为false，后续可以手工修改
        df['is_select'] = False
        df['tag_method'] = 'PE百分位'
        # index_id是人工起的，所以只能作为参数传进来
        df['index_id'] = index_ids
        self._data_helper._upload_basic(df, IndexInfo.__table__.name)

    def index_component(self):
        def _lambda_func_index_component_2(em_id_to_index_id: Dict[str, str], x: str):
            info = json.loads(x)
            r = []
            for one in info:
                em_id = one[0]
                if em_id in ['000915.SH', '000916.SH']:
                    em_id = em_id.split('.')[0] + '.CSI'
                r.append((em_id_to_index_id[em_id], one[1]))
            return json.dumps(r)

        def _lambda_func_index_component_3(wind_id_to_id: Dict[str, str], x: str):
            if x is None:
                return x
            r = []
            info = json.loads(x)
            for one in info:
                try:
                    r.append(wind_id_to_id[one + '.OF'])
                except KeyError:
                    r.append(one + '!0')
            return json.dumps(r)

        index_info_df = self._index_info_df[self._index_info_df.em_id.notna()]
        index_info_df = index_info_df.set_index('em_id')
        em_id_to_index_id: Dict[str, str] = index_info_df['index_id'].to_dict()

        df = self._raw_data_api.get_cs_index_component().drop(columns=['_update_time'])
        df['index_id'] = df['index_id'].map(em_id_to_index_id)
        df['sector'] = df['sector'].transform(lambda x: _lambda_func_index_component_2(em_id_to_index_id, x))
        fund_info_df = self._fund_info_df.set_index('wind_id')
        wind_id_to_id: Dict[str, str] = fund_info_df['fund_id'].to_dict()
        df['related_funds'] = df['related_funds'].transform(lambda x: _lambda_func_index_component_3(wind_id_to_id, x))
        self._data_helper._upload_basic(df, IndexComponent.__table__.name)

    def fund_nav_from_rq(self, start_date, end_date):
        df = self._raw_data_api.get_rq_fund_nav(start_date, end_date)
        df['fund_id'] = df.apply(
            lambda x: self._data_helper._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
        df['subscribe_status'] = df['subscribe_status'].map(self._status_mapping)
        df['redeem_status'] = df['redeem_status'].map(self._status_mapping)
        df = df[df['fund_id'].notna()]
        self._data_helper._upload_basic(df, FundNav.__table__.name)

    def fund_nav(self, start_date, end_date):
        try:
            raw_start_date_dt: datetime.date = pd.to_datetime(start_date, infer_datetime_format=True).date()
            end_date = pd.to_datetime(end_date, infer_datetime_format=True).date()

            # 多取N个交易日的数据
            start_date = raw_start_date_dt - datetime.timedelta(days=40)
            trading_day_df = BasicDataApi().get_trading_day_list(start_date, end_date)
            trading_day_df = trading_day_df[trading_day_df.datetime <= raw_start_date_dt]
            assert trading_day_df.datetime.size >= FUND_NAV_EXTRA_TRADE_DAYS, f'should at least have {FUND_NAV_EXTRA_TRADE_DAYS} trading days'
            # del_date在start_date后边一天，确保可以正确计算ret等daily difference
            start_date = trading_day_df.datetime.array[-FUND_NAV_EXTRA_TRADE_DAYS]
            del_date = trading_day_df.datetime.array[-(FUND_NAV_EXTRA_TRADE_DAYS-1)]
            print(f'[fund_nav] (start_date){start_date} (del_date){del_date} (end_date){end_date}')

            # start_date = start_date.strftime('%Y%m%d')
            df = self._raw_data_api.get_em_fund_nav(start_date, end_date).drop(columns='_update_time')
            df['fund_id'] = df.apply(
                lambda x: self._data_helper._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
            df = df.drop(['CODES'], axis=1)
            df = df[df['fund_id'].notna()]
            df = df.rename(columns={
                    'DATES': 'datetime',
                    'ORIGINALUNIT': 'unit_net_value',
                    'ORIGINALNAVACCUM': 'acc_net_value',
                    'ADJUSTEDNAV': 'adjusted_net_value',
                    'UNITYIELD10K': 'daily_profit',
                    'YIELDOF7DAYS': 'weekly_yield'
                })
            df['weekly_yield'] = df['weekly_yield'].map(lambda x: x / 100.0)

            anv_df = df.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').fillna(method='ffill')
            df = df.set_index(['datetime', 'fund_id'])
            df['change_rate'] = (anv_df / anv_df.shift(1) - 1).stack()

            df = df[df.index.get_level_values(level='datetime').date >= del_date]
            # TODO: 最好应该原子提交下边两步
            BasicDataApi().delete_fund_nav(del_date, end_date)
            self._data_helper._upload_basic(df.reset_index(), FundNav.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_benchmark(self):
        df = pd.read_csv('./data/benchmark_res2.csv')
        df = df[['fund_id', 'fund_em_id', 'benchmark', 'index_text', 'assets', 'industry']].rename(columns={'fund_em_id': 'em_id'})
        self._data_helper._upload_basic(df, FundBenchmark.__table__.name)

    def index_price(self, start_date, end_date):
        try:
            raw_start_date_dt: datetime.date = pd.to_datetime(start_date, infer_datetime_format=True).date()
            # 多取N个交易日的数据
            start_date = raw_start_date_dt - datetime.timedelta(days=90)
            trading_day_df = BasicDataApi().get_trading_day_list(start_date.strftime('%Y%m%d'), end_date)
            trading_day_df = trading_day_df[trading_day_df.datetime <= raw_start_date_dt]
            assert trading_day_df.datetime.size >= INDEX_PRICE_EXTRA_TRADE_DAYS, f'should at least have {INDEX_PRICE_EXTRA_TRADE_DAYS} trading days'
            # del_date在start_date后边一天，确保可以正确计算ret等daily difference
            start_date = trading_day_df.datetime.array[-INDEX_PRICE_EXTRA_TRADE_DAYS]
            del_date = trading_day_df.datetime.array[-(INDEX_PRICE_EXTRA_TRADE_DAYS-1)]
            print(f'[index_price] (start_date){start_date} (del_date){del_date} (end_date){end_date}')
            # 准备多个数据源
            cm_index_price = self._raw_data_api.get_raw_cm_index_price_df(start_date, end_date).drop(columns=['_update_time']).set_index('datetime').fillna(method='ffill')
            cxindex_index_price = self._raw_data_api.get_cxindex_index_price_df(start_date, end_date).drop(columns=['_update_time'])
            yahoo_index_price = self._raw_data_api.get_yahoo_index_price_df(start_date, end_date).drop(columns=['_update_time'])
            index_info_df = self._index_info_df[self._index_info_df.price_source == IndexPriceSource.default]
            em_id_list: Dict[str, str] = index_info_df[['em_id', 'index_id']].dropna().set_index('em_id').to_dict()['index_id']
            em_index_price = self._raw_data_api.get_em_index_price(start_date, end_date, em_id_list.keys()).drop(columns=['_update_time'])
            df_list = []
            # 处理Choice数据
            index_id_list = []
            for em_id, index_id in em_id_list.items():
                df = em_index_price[em_index_price['em_id'] == em_id]
                df = df.drop(columns=['em_id'])
                df['index_id'] = index_id
                df['ret'] = df.close / df.close.shift(1) - 1
                df = df[df.datetime >= pd.to_datetime(del_date).date()]
                df_list.append(df)
                if not df.empty:
                    index_id_list.append(index_id)
                # if em_id == 'SPX.GI':
                #     df['index_id'] = 'sp500'
                #     df_list.append(df.copy())
                #     if not df.empty:
                #         index_id_list.append('sp500')

                #     df = df.set_index('datetime').join(cm_index_price[['usd_central_parity_rate']])
                #     df['index_id'] = 'sp500rmb'
                #     df['open'] *= df['usd_central_parity_rate']
                #     df['close'] *= df['usd_central_parity_rate']
                #     df['low'] *= df['usd_central_parity_rate']
                #     df['high'] *= df['usd_central_parity_rate']
                #     df['ret'] = df.close / df.close.shift(1) - 1
                #     df = df.copy().reset_index().drop('usd_central_parity_rate', axis=1)
                #     df_list.append(df.copy())
                #     if not df.empty:
                #         index_id_list.append('sp500rmb')
                # else:
                #     df_list.append(df)
                #     if not df.empty:
                #         index_id_list.append(index_id)

            end_date_dt: datetime.date = pd.to_datetime(end_date, infer_datetime_format=True).date()
            index_list = ['sp500']  #, 'dax30', 'n225']
            cm_index_list = ['sp500rmb']  #, 'dax30rmb', 'n225rmb']
            for i, c in zip(index_list, cm_index_list):
                df = yahoo_index_price[yahoo_index_price['index_id'] == i]
                df = df.set_index('datetime').reindex(cm_index_price.index).fillna(method='ffill')
                # 存进去人民币计价的数据
                df = df[['close', 'open', 'low', 'high', 'volume']]
                df['close'] *= cm_index_price['usd_central_parity_rate']
                df['open'] *= cm_index_price['usd_central_parity_rate']
                df['low'] *= cm_index_price['usd_central_parity_rate']
                df['high'] *= cm_index_price['usd_central_parity_rate']
                df['index_id'] = c
                df['ret'] = df.close / df.close.shift(1) - 1
                df = df.loc[raw_start_date_dt:end_date_dt, :]
                df_list.append(df.reset_index())

            df_list.append(cxindex_index_price[cxindex_index_price.datetime == end_date_dt])
            # 存进去美元计价的数据
            df_list.append(yahoo_index_price[yahoo_index_price.datetime.between(raw_start_date_dt, end_date_dt)])

            df = pd.concat(df_list, ignore_index=True)
            # TODO: 改成原子提交下边两个操作
            BasicDataApi().delete_index_price(index_id_list, del_date, end_date)
            self._data_helper._upload_basic(df.replace([np.Inf, -np.Inf], np.nan), IndexPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_rating_latest(self):
        try:
            fund_rating_df = self._raw_data_api.get_fund_rating()
            fund_info = self._fund_info_df[['fund_id', 'order_book_id', 'start_date', 'end_date']]
            score_name = ['zs', 'sh3', 'sh5', 'jajx']
            score_dic = {}
            for s in score_name:
                score_df = fund_rating_df[['order_book_id', 'datetime', s]]
                datelist = sorted(list(set(score_df.dropna().datetime.tolist())))
                score_dic[s] = datelist[-1]
            res = []
            for s, d in score_dic.items():
                try:
                    df = fund_rating_df[['order_book_id', 'datetime', s]]
                    df = df[df['datetime'] == d]
                    con1 = fund_info['start_date'] <= d
                    con2 = fund_info['end_date'] >= d
                    fund_info_i = fund_info[con1 & con2]
                    dic = {row['order_book_id']: row['fund_id']
                        for index, row in fund_info_i.iterrows()}
                    df['fund_id'] = df['order_book_id'].map(lambda x: dic[x])
                    df = df[['fund_id', s]].copy().set_index('fund_id')
                    res.append(df)
                except Exception as e:
                    print(e)
            df = pd.concat(res, axis=1, sort=False)
            df['fund_id'] = df.index
            df['update_time'] = datetime.date.today()
            self._data_helper._upload_basic(df, FundRatingLatest.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_stock_price(self, start_date, end_date):
        try:
            stock_price_df = self._raw_data_api.get_rq_stock_price(start_date, end_date)
            stock_post_price_df = self._raw_data_api.get_rq_stock_post_price(start_date, end_date)

            stock_post_price_df['adj_close'] = stock_post_price_df['close']
            stock_post_price_df = stock_post_price_df.filter(items=['adj_close', 'datetime', 'order_book_id'],
                axis='columns')

            stock_price_merge_df = stock_price_df.merge(stock_post_price_df, how='left', on=['datetime', 'order_book_id'])
            stock_price_merge_df['post_adj_factor'] = stock_price_merge_df['adj_close'] / stock_price_merge_df['close']
            stock_price_merge_df = stock_price_merge_df.rename(columns={'order_book_id': 'stock_id'})
            self._data_helper._upload_basic(stock_price_merge_df, StockPrice.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def rq_fund_ret(self, start_date, end_date):
        try:
            column_rename_dict = {
                'sharp_ratio': 'sharpe_ratio',
                'last_week_return': 'w1_ret',
                'last_month_return': 'm1_ret',
                'last_three_month_return': 'm3_ret',
                'last_six_month_return': 'm6_ret',
                'last_twelve_month_return': 'y1_ret',
                'max_drop_down': 'mdd',
                'annualized_returns': 'annual_ret',
                'average_size': 'avg_size',
                'information_ratio': 'info_ratio',
                'to_date_return':'to_date_ret'
            }
            column_drop_list = ['order_book_id', 'year_to_date_return', 'annualized_risk']

            df = self._raw_data_api.get_rq_fund_indicator(start_date, end_date)
            df['fund_id'] = df.apply(
                lambda x: self._data_helper._get_fund_id_from_order_book_id(x['order_book_id'], x['datetime']), axis=1)
            df = df.rename(columns=column_rename_dict).drop(columns=column_drop_list)
            df = df[df['fund_id'].notnull()]

            self._data_helper._upload_basic(df, FundRet.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def stock_info(self):
        try:
            df = self._raw_data_api.get_stock_info()
            self._data_helper._upload_basic(df, StockInfo.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def sector_info(self):
        try:
            df = pd.read_csv('./data/sector_info.csv', index_col=0)
            df['sector_type'] = df['sector_type'].map({'industry': SectorType.industry.name, 'topic': SectorType.topic.name})
            self._data_helper._upload_basic(df, SectorInfo.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def sector_funds(self):
        def _lambda_sector_funds_1(em_id_map: Dict[str, str], x: str):
            try:
                return em_id_map[x]
            except KeyError:
                return x

        try:
            index_info_df = self._index_info_df[self._index_info_df.em_id.notna()]
            index_info_df = index_info_df.set_index('em_id')
            em_id_list: List[str] = index_info_df.index.to_list()
            em_id_to_index_id: Dict[str, str] = index_info_df['index_id'].to_dict()

            df = pd.read_csv('./data/sector_funds.csv', index_col=0)
            df['csi_index_id'] = df['csi_index_id'].transform(lambda x: _lambda_sector_funds_1({em_id.split('.')[0]: em_id for em_id in em_id_list if em_id is not None}, x))
            df['index_id'] = df['csi_index_id'].map(em_id_to_index_id)
            self._data_helper._upload_basic(df, SectorFunds.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_size(self, start_date, end_date):
        try:
            # Since fund_size in basic db is latest snapshot, we only use end_date as param
            df = self._raw_data_api.get_em_fund_scale(end_date=end_date)
            if df is None:
                print('failed to get em fund scale for fund size')
                return False
            # 将场内变为场外的
            df['CODES'] = df.CODES.transform(lambda x: x.split('.')[0] + '.OF' if '.OF' not in x else x)
            df = df.pivot_table(index='DATES', columns='CODES', values='FUNDSCALE').fillna(method='ffill')
            df = df.iloc[[-1], :].apply(lambda x: pd.Series({'fund_id': self._data_helper._get_fund_id_from_order_book_id(x.name.split('.')[0], x.index.array[0]), 'latest_size': float(x.array[0])})).T
            df = df.reset_index(drop=True)
            df = df[df.notnull().all(axis=1)]
            self._data_helper._upload_basic(df, FundSize.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_status_latest(self, end_date: str):
        try:
            df: Optional[pd.DataFrame] = self._raw_data_api.get_em_fund_status(end_date=end_date)
            if df is None:
                print('failed to get em fund scale for fund status latest')
                return False
            # 将场内变为场外的
            df['CODES'] = df.CODES.transform(lambda x: x.split('.')[0] + '.OF' if '.OF' not in x else x)
            df = df.pivot(index='DATES', columns='CODES', values=['PURCHSTATUS', 'REDEMSTATUS', 'LPLIMIT', 'OTCLPLIMITJG', 'OTCLPLIMITGR', 'TRADESTATUS']).fillna(method='ffill')
            df = df.iloc[[-1], :].stack().reset_index()
            df['fund_id'] = df[['DATES', 'CODES']].apply(lambda x: self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
            df = df[df.fund_id.notna()]
            df = df.drop(columns=['DATES', 'CODES']).rename(columns=lambda x: FundStatusLatest.__getattribute__(FundStatusLatest, x).name)
            self._data_helper._upload_basic(df, FundStatusLatest.__table__.name, to_truncate=True)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_rate(self, end_date):
        def _count_star_history(x):
            color_change = {
                '红':'蓝',
                '蓝':'白',
                '白':'红',
            }
            if x is None:
                return None
            if '星' in x:
                return str(x.count('★') + x.count('☆') + 1) + color_change[x[-2]]
            else:
                return str(x.count('★') + x.count('☆'))

        def _count_A_rating(x: Optional[str]) -> Optional[str]:
            if x is None:
                return
            return str(x.count('A'))

        def _count_star_recent(x):
            if x is None:
                return None
            if '星' in x:
                return str(x.count('★') + x.count('☆') + 1) + x[-2]
            else:
                return str(x.count('★') + x.count('☆'))
        try:
            df = self._raw_data_api.get_em_fund_rate(end_date)
            df['fund_id'] = df.apply( lambda x : self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
            no_star_columns = ['CODES','DATES','_update_time','fund_id']
            A_rate_columns = ['TXCOMRATRAT']
            for col in df:
                if col in no_star_columns:
                    continue
                if col in A_rate_columns:
                    df[col] = df[col].map(_count_A_rating)
                else:
                    df[col] = df[col].map(lambda x: _count_star_recent(x))
            df = df.drop(['CODES','_update_time'], axis=1)
            name_dic = {
                'DATES':'datetime',
                'MKTCOMPRE3YRAT':'mkt_3y_comp',
                'MORNSTAR3YRAT':'mornstar_3y',
                'MORNSTAR5YRAT':'mornstar_5y',
                'MERCHANTS3YRAT':'merchant_3y',
                'TXCOMRATRAT':'tx_comp',
                'SHSTOCKSTAR3YCOMRAT':'sh_3y_comp',
                'SHSTOCKSTAR3YSTOCKRAT':'sh_3y_stock',
                'SHSTOCKSTAR3YTIMERAT':'sh_3y_timeret',
                'SHSTOCKSTAR3YSHARPERAT':'sh_3y_sharpe',
                'SHSTOCKSTAR5YCOMRATRAT':'sh_5y_comp',
                'SHSTOCKSTAR5YSTOCKRAT':'sh_5y_stock',
                'SHSTOCKSTAR5YTIMERAT':'sh_5y_timeret',
                'SHSTOCKSTAR5YSHARPRAT':'sh_5y_sharpe',
                'JAJXCOMRAT':'jian_comp',
                'JAJXEARNINGPOWERRAT':'jian_earn',
                'JAJXACHIEVESTABILITYRAT':'jian_stable',
                'JAJXANTIRISKRAT':'jian_risk',
                'JAJXSTOCKSELECTIONRAT':'jian_stock',
                'JAJXTIMESELECTIONRAT':'jian_timeret',
                'JAJXBENCHMARKTRACKINGRAT':'jian_track',
                'JAJXEXCESSEARNINGSRAT':'jian_alpha',
                'JAJXTOTALFEERAT':'jian_fee',
                'fund_id':'fund_id',
            }
            df = df.rename(columns = name_dic)
            df = df.dropna(subset=['fund_id','datetime'], axis=0)
            # 寻找离目前最近且没超过当前时间的财报日期，将数据写在这个日期上（先删掉这个日期上的旧数据）
            end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True)
            md = end_date_dt.strftime('%m%d')
            sentry_date = pd.Series(self._REPORT_DATE_LIST)
            sentry = sentry_date[sentry_date <= md]
            if sentry.empty:
                # 这里取前一年的12月31
                real_date = datetime.date(end_date_dt.year - 1, 12, 31)
            else:
                real_date = pd.to_datetime(str(end_date_dt.year)+sentry.array[-1], infer_datetime_format=True).date()
            df['datetime'] = real_date
            # TODO: 最好原子提交下边两步
            BasicDataApi().delete_fund_rate(real_date, df.fund_id.to_list())
            self._data_helper._upload_basic(df, FundRate.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def trading_day_list(self, start_date, end_date, is_automatic_update=True):
        try:
            # 如果是每日自动更新，我们需要去存T+1日的数据
            if is_automatic_update:
                start_date = end_date
                end_date = ''
            df = self._raw_data_api.get_em_tradedates(start_date, end_date)
            df = df.rename(columns={'TRADEDATES': 'datetime'})
            if is_automatic_update:
                df = df.iloc[[1], :]
            self._data_helper._upload_basic(df, TradingDayList.__table__.name)
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def fund_size_hold_rate_history(self):
        start_date = '20200630'
        end_date = '20200630'

        fund_holding_rate = self._raw_data_api.get_em_fund_holding_rate(start_date=start_date, end_date=end_date)
        fund_holding_rate['fund_id'] = fund_holding_rate.apply(
            lambda x: self._data_helper._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
        fund_holding_rate = fund_holding_rate[fund_holding_rate.fund_id.notna()]
        fund_holding_rate = fund_holding_rate.drop(['CODES', '_update_time'], axis=1).rename(columns={
            'DATES': 'datetime',
            'HOLDPERSONALHOLDINGPCT': 'personal_holds',
            'HOLDINSTIHOLDINGPCT': 'institution_holds',
            'HOLDNUM': 'hold_num',
        })
        fund_holding_rate = fund_holding_rate[fund_holding_rate.drop(columns='fund_id').notna().any(axis=1)]

        fund_scale = self._raw_data_api.get_em_fund_scale(start_date=start_date, end_date=end_date)
        fund_scale = fund_scale.pivot(index='DATES', columns='CODES', values='FUNDSCALE').fillna(method='ffill').stack()
        fund_scale = pd.DataFrame(fund_scale, columns=['FUNDSCALE']).reset_index()
        fund_scale['fund_id'] = fund_scale.apply(
            lambda x: self._data_helper._get_fund_id_from_order_book_id(x['CODES'].split('.')[0], x['DATES']), axis=1)
        fund_scale = fund_scale[fund_scale.fund_id.notna()]
        fund_scale = fund_scale.drop(['CODES'], axis=1).rename(columns={'DATES': 'datetime', 'FUNDSCALE': 'size'})
        fund_scale = fund_scale.dropna().drop_duplicates(subset=['size', 'fund_id'], keep='first')

        df = fund_scale.set_index(['fund_id', 'datetime']).join(fund_holding_rate.set_index(['fund_id', 'datetime']), how='outer').reset_index()
        self._data_helper._upload_basic(df, Fund_size_and_hold_rate.__table__.name)

    def fund_hold_asset(self, start_date: str, end_date: str):
        # history, update only on report date
        em_hold_asset = self._raw_data_api.get_em_fund_hold_asset(start_date, end_date)
        em_hold_asset['CODES'] = em_hold_asset.apply(lambda x:
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_asset = em_hold_asset.dropna(subset=['CODES']).drop(columns='_update_time')
        name_dict = {
            'CODES': 'fund_id',
            'DATES': 'datetime',
            'PRTSTOCKTONAV': 'stock_nav_ratio',
            'PRTBONDTONAV': 'bond_nav_ratio',
            'PRTCASHTONAV': 'cash_nav_ratio',
            'PRTOTHERTONAV': 'other_nav_ratio',
            'MMFFIRSTREPOTONAV': 'first_repo_to_nav',
            'MMFAVGPTM': 'avg_ptm',
        }
        em_hold_asset = em_hold_asset.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_asset, FundHoldAsset.__table__.name)

    def fund_hold_industry(self, end_date: str):
        # history, update only on report date
        em_hold_industry = self._raw_data_api.get_em_fund_hold_industry(end_date)
        em_hold_industry['CODES'] = em_hold_industry.apply(lambda x:
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_industry = em_hold_industry.dropna(subset=['CODES']).drop(columns='_update_time')
        name_dict = {
            'CODES': 'fund_id',
            'DATES': 'datetime',
        }
        em_hold_industry = em_hold_industry.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_industry, FundHoldIndustry.__table__.name)

    def fund_hold_stock(self, end_date: str):
        #history, update only on report date
        em_hold_stock = self._raw_data_api.get_em_fund_hold_stock(end_date)
        em_hold_stock['CODES'] = em_hold_stock.apply(lambda x :
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_stock = em_hold_stock.dropna(subset=['CODES']).drop('_update_time', axis=1)
        name_dict = {
            'CODES':'fund_id',
            'DATES':'datetime',
        }
        em_hold_stock = em_hold_stock.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_stock, FundHoldStock.__table__.name)

    def fund_hold_bond(self, end_date: str):
        #history, update only on report date
        em_hold_bond = self._raw_data_api.get_em_fund_hold_bond(end_date)
        em_hold_bond['CODES'] = em_hold_bond.apply(lambda x :
                            self._data_helper._get_fund_id_from_order_book_id(x.CODES.split('.')[0], x.DATES), axis=1)
        em_hold_bond = em_hold_bond.dropna(subset=['CODES']).drop('_update_time', axis=1)
        name_dict = {
            'CODES':'fund_id',
            'DATES':'datetime',
        }
        em_hold_bond = em_hold_bond.rename(columns=name_dict)
        self._data_helper._upload_basic(em_hold_bond, FundHoldBond.__table__.name)

    def mng_info(self):
        df = self._raw_data_api.get_em_mng_info()
        df['fund_id'] = df.apply(lambda x : 
                    self._data_helper._get_fund_id_from_order_book_id(x.code.split('.')[0], datetime.date(2020,8,10)), axis=1)
        df = df.drop('code',axis=1).dropna(subset=['fund_id'])
        self._data_helper._upload_basic(df, MngInfo.__table__.name)

    def mng_change(self):
        df = self._raw_data_api.get_em_fund_mng_change()
        df['fund_id'] = df.apply(lambda x : 
                    self._data_helper._get_fund_id_from_order_book_id(x.code.split('.')[0], datetime.date(2020,8,10)), axis=1)
        df = df.drop('code',axis=1).dropna(subset=['fund_id'])
        self._data_helper._upload_basic(df, FundMngChange.__table__.name)

    def fund_comp_mng_change(self):
        df = self._raw_data_api.get_fund_com_mng_change()
        self._data_helper._upload_basic(df, FundCompMngChange.__table__.name)

    def fund_comp_core_mng(self):
        df = self._raw_data_api.get_em_fund_com_core_mng()
        df['fund_id'] = df.apply(lambda x : 
                    self._data_helper._get_fund_id_from_order_book_id(x.code.split('.')[0], datetime.date(2020,8,10)), axis=1)
        df = df.drop('code',axis=1).dropna(subset=['fund_id'])
        self._data_helper._upload_basic(df, FundCompCoreMng.__table__.name)

    def fund_open_info(self):
        df = self._raw_data_api.get_em_fund_open()
        df['fund_id'] = df.apply(
            lambda x : self._data_helper._get_fund_id_from_order_book_id(x.em_id.split('.')[0], datetime.date(2020,9,15)), axis = 1)       
        df = df.dropna(subset=['fund_id'])
        df = df.drop(['em_id','_update_time'],axis=1)
        self._data_helper._upload_basic(df, FundOpenInfo.__table__.name)

    def process_all(self, start_date, end_date):
        failed_tasks = []
        # Depend on RQ
        # if not self.rq_fund_ret(start_date, end_date):
        #     failed_tasks.append('rq_fund_ret')

        # # Depend on RQ
        # if not self.stock_info():
        #     failed_tasks.append('stock_info')

        # Depend on RQ
        # if not self.rq_stock_price(start_date, end_date):
        #     failed_tasks.append('rq_stock_price')

        if not self.index_price(start_date, end_date):
            failed_tasks.append('index_price')

        # new_fund_list, _ = RawDataHelper.get_new_and_delisted_fund_list(end_date)
        # if new_fund_list is None:
        #     failed_tasks.append('get new_fund_list')
        # else:
        #     if not self.fund_info_from_raw(new_fund_list):
        #         failed_tasks.append('fund_info')

        # TODO: Depend on RQ, change to EM
        if not self.fund_rating_latest():
            failed_tasks.append('fund_rating_latest')

        if not self.fund_size(start_date, end_date):
            failed_tasks.append('fund_size')

        if not self.fund_nav(start_date, end_date):
            failed_tasks.append('fund_nav')

        if not self.trading_day_list(start_date, end_date):
            failed_tasks.append('trading_day_list')

        if not self.fund_rate(end_date):
            failed_tasks.append('fund_rate')

        if not self.fund_status_latest(end_date):
            failed_tasks.append('fund_status_latest')

        return failed_tasks

if __name__ == '__main__':
    bdp = BasicDataPart1(BasicDataHelper())
    # start_date = '20200428'
    # end_date = '20200428'
    # bdp.fund_nav(start_date, end_date)
    # bdp.index_price(start_date, end_date)
    # bdp.rq_stock_price(start_date, end_date)
    # bdp.rq_fund_ret(start_date, end_date)
    # bdp.fund_rating_latest()
    # bdp.fund_size('20200513', '20200513')
    # bdp.trading_day_list('20200124', '20200301', False)
    # bdp.fund_benchmark()
    # bdp.sector_info()
    # bdp.sector_funds()
    # bdp.index_component()
    # bdp.fund_rate('20200812')
    # bdp.fund_size_hold_rate_history()



from cassandra.cluster import Cluster
from cassandra.cqlengine import connection
from cassandra.auth import PlainTextAuthProvider
from ....util.config import SurfingConfigurator
from ....util.singleton import Singleton


class CassandraConnector(metaclass=Singleton):

    def __init__(self):
        self.conf = SurfingConfigurator().get_cassandra_setting()
        self.session = None
        self.connection = None

    def init_conn(self):
        if not self.session or not self.connection:
            auth_provider = PlainTextAuthProvider(
                username=self.conf.username,
                password=self.conf.password,
            )
            cluster = Cluster(
                self.conf.hosts,
                auth_provider=auth_provider,
            )
            self.session = cluster.connect()
            self.connection = connection.register_connection('surfing', session=self.session)
            # self.connection = connection.register_connection('', session=self.session, default=True)

    def get_conn(self):
        self.init_conn()
        return self.connection

    def get_session(self):
        self.init_conn()
        return self.session



import numpy as np
import pandas as pd
import datetime
import traceback
from ...api.basic import BasicDataApi
from ...api.raw import RawDataApi
from ...api.derived import DerivedDataApi
from ...view.derived_models import StyleBox
from .derived_data_helper import DerivedDataHelper


class StyleBoxGenerator:

    _REPORT_DATE_LIST = ['0331', '0630', '0930', '1231']

    def __init__(self, _data_helper: DerivedDataHelper):
        self._raw_api = RawDataApi()
        self._basic_api = BasicDataApi()
        self._data_helper: DerivedDataHelper = _data_helper

    def rank_score(self, a):
        a_rank = a.rank()
        a_score = 100 * (a_rank - a_rank.min()) / (a_rank.max() - a_rank.min())
        return a_score

    def init(self, start_date, end_date):
        # 导入并格式化股价
        self.stock_price_raw = self._raw_api.get_em_stock_price(start_date, end_date, columns=['close']).set_index(['stock_id', 'datetime'])
        self.stock_price = self.stock_price_raw.reset_index().pivot(index='datetime', columns='stock_id', values='close')
        # 导入并格式化总股本
        self.stock_daily_info_raw = self._raw_api.get_em_daily_info(start_date, end_date, columns=['total_share']).set_index(['stock_id', 'datetime'])
        self.stock_daily_info = self.stock_daily_info_raw.reset_index().pivot(index='datetime', columns='stock_id', values='total_share')
        # 相乘计算出总市值
        self.cap = (self.stock_price * self.stock_daily_info).iloc[0]
        # 导入并格式化财务数据
        start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
        start_date_2 = datetime.date(start_date_dt.year - 2, start_date_dt.month, start_date_dt.day)
        self.fin_fac = self._raw_api.get_em_stock_fin_fac(start_date=start_date_2, end_date=end_date)
        self.earnings = self.fin_fac.reset_index().pivot(index='datetime', columns='stock_id', values='income_statement_61')
        self.book_values = self.fin_fac.reset_index().pivot(index='datetime', columns='stock_id', values='balance_statement_74')
        self.revenues = self.fin_fac.reset_index().pivot(index='datetime', columns='stock_id', values='income_statement_9')
        self.cash = self.fin_fac.reset_index().pivot(index='datetime', columns='stock_id', values='cashflow_statement_39')
        # 计算所需的数据
        self.earnings_td = self.earnings.iloc[-1].dropna()
        self.bv_td = self.book_values.iloc[-1].dropna()
        self.rev_td = self.revenues.iloc[-1].dropna()
        self.cash_td = self.cash.iloc[-1].dropna()
        # 计算OGS打分所需的次级指标
        self.erate = self.earnings.pct_change().mean().dropna()
        self.brate = self.book_values.pct_change().mean().dropna()
        self.rrate = self.revenues.pct_change().mean().dropna()
        self.crate = self.cash.pct_change().mean().dropna()
        self.ers = self.rank_score(self.erate)
        self.brs = self.rank_score(self.brate)
        self.rrs = self.rank_score(self.rrate)
        self.crs = self.rank_score(self.crate)
        # 去掉计算所得数据为NaN的股票
        self.lis = self.earnings_td.index & self.cap.index & self.bv_td.index & self.rev_td.index & self.cash_td.index & self.erate.index & self.brate.index & self.rrate.index & self.crate.index
        # 计算OVS打分所需的次级指标
        self.ep = self.earnings_td[self.lis] / self.cap[self.lis]
        self.bp = self.bv_td[self.lis] / self.cap[self.lis]
        self.rp = self.rev_td[self.lis] / self.cap[self.lis]
        self.cp = self.cash_td[self.lis] / self.cap[self.lis]
        self.eps = self.rank_score(self.ep)
        self.bps = self.rank_score(self.bp)
        self.rps = self.rank_score(self.rp)
        self.cps = self.rank_score(self.cp)
        # 加权计算OVS和OGS
        self.ovs = 0.5 * self.eps + (self.bps + self.rps + self.cps) / 6
        self.ogs = (self.ers + self.brs + self.rrs + self.crs) / 4
        # 得到Value Core Growth
        self.vcg = self.ogs - self.ovs
        # 再次取交集筛去NaN
        self.lis2 = self.cap.dropna().index & self.vcg.dropna().index
        self.cap = self.cap[self.lis2]
        self.vcg = self.vcg[self.lis2]

    def cal_stock(self):
        self.capsum = self.cap.sort_values(ascending=False).cumsum()
        self.capt = self.capsum / self.capsum[-1]
        self.big_cap = self.capt[self.capt < 0.7].index
        self.mid_cap = self.capt[np.logical_and(self.capt >= 0.7, self.capt <= 0.9)].index
        self.small_cap = self.capt[self.capt > 0.9].index
        self.big_vcg = self.vcg[self.big_cap]
        self.mid_vcg = self.vcg[self.mid_cap]
        self.small_vcg = self.vcg[self.small_cap]
        self.LMT = self.cap[self.mid_cap[0]]
        self.MST = self.cap[self.mid_cap[-1]]
        x = []
        y = []
        d = []
        for i in range(len(self.vcg)):
            idx = self.vcg.index[i]
            ytemp = 100 * (1 + (np.log(self.cap.loc[idx]) - np.log(self.MST)) / (np.log(self.LMT) - np.log(self.MST)))
            y.append(ytemp)
            if ytemp > 200:
                xtemp = 100 * (1 + (self.vcg.iloc[i] - self.big_vcg.quantile(1 / 3)) / (self.big_vcg.quantile(2 / 3) - self.big_vcg.quantile(1 / 3)))
            elif ytemp < 100:
                xtemp = 100 * (1 + (self.vcg.iloc[i] - self.small_vcg.quantile(1 / 3)) / (self.small_vcg.quantile(2 / 3) - self.small_vcg.quantile(1 / 3)))
            else:
                xtemp = 100 * (1 + (self.vcg.iloc[i] - self.mid_vcg.quantile(1 / 3)) / (self.mid_vcg.quantile(2 / 3) - self.mid_vcg.quantile(1 / 3)))
            x.append(xtemp)
            d.append(idx)
        xdf = pd.DataFrame(x, index=d)
        ydf = pd.DataFrame(y, index=d)
        self.stock_res = pd.concat([xdf, ydf], axis=1)
        self.stock_res.columns = ['x', 'y']

    # repo_date数据类型为datetime.date
    def cal_fund(self, repo_date):
        fund_list = self._basic_api.get_fund_info()
        fund_list = fund_list[np.logical_or(fund_list.wind_class_1 == '股票型基金', fund_list.wind_class_2 == '偏股混合型基金')]
        fund_list = fund_list[fund_list.end_date > repo_date]
        self.flist = fund_list[['wind_id', 'fund_id']].set_index('wind_id').to_dict()['fund_id']
        xres = []
        yres = []
        dres = []
        dtemp = []
        repo_str = repo_date.strftime('%Y%m%d')
        temp_list = self._raw_api.get_wind_fund_stock_portfolio(start_date=repo_str, end_date=repo_str, wind_fund_list=list(self.flist.keys()))
        for item, fund_id in self.flist.items():
            # print(item+ ': ' +str(c)+'/'+str(l))
            temp = temp_list[temp_list.S_INFO_WINDCODE == item]
            if len(temp) == 0:
                # print(item + ' No Fund Data')
                continue
            temp_data = temp[['S_INFO_STOCKWINDCODE', 'F_PRT_STKVALUE']]
            temp_data.columns = ['stock_id', 'value']
            temp_data = temp_data.reset_index()
            temp_data['weight'] = temp_data['value'] / (temp_data['value'].sum())
            xf = 0
            yf = 0
            wf = 0
            for i in range(len(temp_data)):
                idx = temp_data.stock_id[i]
                if idx in self.stock_res.index:
                    wf += temp_data.weight[i]
                    xf += temp_data.weight[i] * self.stock_res.loc[idx].x
                    yf += temp_data.weight[i] * self.stock_res.loc[idx].y
                else:
                    # print('Error, Stock Data Missing: ' + idx)
                    continue
            if wf <= 0.5:
                # print(item + 'We do not have enough stock data')
                continue

            xf = xf / wf
            yf = yf / wf
            xres.append(xf)
            yres.append(yf)
            dres.append(fund_id)
            dtemp.append(repo_str)
        xres_df = pd.DataFrame(xres, index=dres)
        yres_df = pd.DataFrame(yres, index=dres)
        dtemp = pd.DataFrame(dtemp, index=dres)
        # print(xres_df, yres_df, dtemp)
        self.fund_res = pd.concat([xres_df, yres_df], axis=1)
        self.fund_res = pd.concat([dtemp, self.fund_res], axis=1)
        self.fund_res = self.fund_res.reset_index()
        self.fund_res.columns = ['fund_id', 'datetime', 'x', 'y']

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True)
            md = end_date_dt.strftime('%m%d')
            sentry_date = pd.Series(self._REPORT_DATE_LIST)
            sentry = sentry_date[sentry_date <= md]
            if sentry.empty:
                # 这里取前一年的12月31
                real_date = datetime.date(end_date_dt.year - 1, 12, 31)
            else:
                real_date = pd.to_datetime(str(end_date_dt.year) + sentry.array[-1], infer_datetime_format=True).date()
            self.init(start_date, end_date)
            self.cal_stock()
            self.cal_fund(real_date)
            # # 获取当前的style box
            # now_df = DerivedDataApi().get_fund_style_box(datetime=real_date)
            # if now_df is not None and not now_df.empty:
            #     # 过滤出来新增基金style box以及旧基金的style box发生变化的情况
            #     self.fund_res = self.fund_res.round(12).merge(now_df.drop(columns=['_update_time']), how='left', on=['fund_id', 'datetime', 'x', 'y'], indicator=True, validate='one_to_one')
            #     self.fund_res = self.fund_res[self.fund_res._merge == 'left_only'].drop(columns=['_merge'])
            # TODO: 最好原子提交下边两步
            # DerivedDataApi().delete_fund_style_box(real_date, self.fund_res.fund_id.to_list())
            self.fund_res['datetime'] = end_date
            self._data_helper._upload_derived(self.fund_res, StyleBox.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('style_box')
        return failed_tasks


if __name__ == "__main__":
    date = '20200824'
    sbg = StyleBoxGenerator(DerivedDataHelper())
    sbg.process(date, date)



from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from ...wrapper.cas.cas_connection import CassandraConnector
from ...wrapper.cas.constants import CAS_KEY_SPACE

CassandraConnector().get_conn()


class ThirdPartyPortfolioNav(Model):

    __table_name__ = 'third_party_portfolio_nav'
    __keyspace__ = CAS_KEY_SPACE
    __connection__ = 'surfing'

    po_id = columns.Text(primary_key=True)  # 组合ID
    date = columns.List(columns.Text)  # 日期
    nav = columns.List(columns.Double)  # 组合净值

    def to_dict(self):
        return {
            'po_id': self.po_id,
            'date': self.date,
            'nav': self.nav,
        }




import datetime
from typing import List, Tuple, Optional
import pandas as pd
import traceback

from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...view.derived_models import FundScoreExtended
from .derived_data_helper import DerivedDataHelper, normalize, score_rescale


class FundScoreGroupV1Process:

    _SPECIAL_LIST = ['国际(QDII)基金']
    _BLOCK_LIST = ('商品型基金',)

    def __init__(self, data_helper: DerivedDataHelper, data_cycle: int):
        self._data_helper = data_helper
        self._data_cycle = str(data_cycle) + 'Y'
        self._result: Optional[pd.DataFrame] = None

    def init(self, start_date: str, end_date: str):
        self._end_date = end_date
        self._fund_info: pd.DataFrame = BasicDataApi().get_fund_info().drop(columns='_update_time').set_index('fund_id')
        # 获取计算评分需要用到的indicator
        self._fund_indicator_group: pd.DataFrame = DerivedDataApi().get_fund_indicator_group(start_date, end_date, self._data_cycle).drop(columns='_update_time').set_index('fund_id')

    @staticmethod
    def _get_columns_weight_by_wind_class(wind_class: str) -> Tuple[List[str], List[float]]:
        if wind_class in ('偏股混合型基金', '普通股票型基金', '平衡混合型基金', '灵活配置型基金','股票多空'):
            columns_weight = {'annual_ret': 0.2, 'sharpe': 0.6, 'info_ratio': 0.15, 'continue_stats_v': 0.05}
        elif wind_class in ('中长期纯债型基金', '短期纯债型基金', '混合债券型一级基金', '混合债券型二级基金', '偏债混合型基金', '货币市场型基金'):
            columns_weight = {'annual_ret': 0.3, 'sharpe': 0.6, 'continue_stats_v': 0.1}
        elif wind_class in ('增强指数型基金', '增强指数型债券基金'):
            columns_weight = {'annual_ret': 0.2, 'continue_stats_v': 0.05, 'alpha': 0.75}
        elif wind_class in ('被动指数型基金', '被动指数型债券基金'):
            columns_weight = {'annual_ret': 0.2, 'continue_stats_v': 0.1, 'treynor': 0.7}
        elif wind_class in ('国际(QDII)基金', 'REITs'):
            columns_weight = {'annual_ret': 0.2, 'sharpe': 0.175, 'info_ratio': 0.175, 'continue_stats_v': 0.1, 'alpha': 0.175, 'treynor': 0.175}
        else:
            assert False, f'invalid wind class!! {wind_class}'
        return list(columns_weight.keys()), list(columns_weight.values())

    def _calc_score(self, x: pd.DataFrame, by_key: str) -> pd.DataFrame:
        fin_data: pd.DataFrame = self._fund_indicator_group.loc[x.index.intersection(self._fund_indicator_group.index), :]
        wind_class: str = x[by_key].array[0]

        if wind_class in ('偏股混合型基金', '普通股票型基金', '平衡混合型基金', '灵活配置型基金'):
            data_to_calc_ret: pd.DataFrame = fin_data[fin_data.sharpe >= -20]
        else:
            data_to_calc_ret: pd.DataFrame = fin_data[fin_data.alpha <= 20]
        if data_to_calc_ret.empty:
            return
        columns, weight = FundScoreGroupV1Process._get_columns_weight_by_wind_class(wind_class)
        result1 = pd.DataFrame(index=data_to_calc_ret.index)
        result1['return_score'] = score_rescale((weight * normalize(data_to_calc_ret[columns])).sum(axis=1, keepdims=True))

        data_to_calc_others: pd.DataFrame = fin_data[fin_data.sharpe >= -20]
        result2 = pd.DataFrame(index=data_to_calc_others.index)
        result2['robust_score'] = score_rescale(-normalize(data_to_calc_others[['annual_vol']]))

        if wind_class in ('货币市场型基金'):
            # 这里是为了配合db，实际应该是scale_score
            result2['allocation_score'] = score_rescale(normalize(data_to_calc_others[['scale']]))
        else:
            result2['timing_score'] = score_rescale(normalize(data_to_calc_others[['stock_cl_beta']]))
            result2['selection_score'] = score_rescale(normalize(data_to_calc_others[['stock_cl_alpha']]))
            # 这里是为了配合db，实际应该是risk_score
            result2['allocation_score'] = score_rescale(-normalize(data_to_calc_others[['mdd']]))

        # 合并所有的评分并计算总分
        total = result1.join(result2, how='outer')
        total['total_score'] = score_rescale(normalize(total.sum(axis=1).to_frame()))
        return total

    def calc(self):
        # 按wind class 2计算
        df1: pd.DataFrame = self._fund_info[(~self._fund_info.wind_class_1.isin(self._SPECIAL_LIST)) & (~self._fund_info.wind_class_2.isin(self._BLOCK_LIST))].groupby(by='wind_class_2', sort=False).apply(self._calc_score, by_key='wind_class_2')
        if not df1.empty:
            df1 = df1.droplevel(level=0)
            df1['wind_class_1'] = self._fund_info.loc[df1.index, 'wind_class_1']
        # 按wind class 1计算
        df2: pd.DataFrame = self._fund_info[self._fund_info.wind_class_1.isin(self._SPECIAL_LIST)].groupby(by='wind_class_1', sort=False).apply(self._calc_score, by_key='wind_class_1')
        if not df2.empty:
            df2 = df2.reset_index().set_index('fund_id')
        # 两部分连到一起
        self._result = df1.append(df2)
        if not self._result.empty:
            self._result['datetime'] = self._end_date
            self._result['data_cycle'] = self._data_cycle

    def process(self, start_date: str, end_date: str, end_date_dt: datetime.date) -> List[str]:
        print(f'fund score group update on the last day of week {end_date_dt}')
        failed_tasks = []
        try:
            self.init(start_date, end_date)
            self.calc()
            if not self._result.empty:
                self._data_helper._upload_derived(self._result.reset_index(), FundScoreExtended.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_score_group')
        return failed_tasks

    @staticmethod
    def update_history(end_date: str, data_cycle: int):
        from ...wrapper.mysql import BasicDatabaseConnector
        from ...view.basic_models import TradingDayList
        with BasicDatabaseConnector().managed_session() as quant_session:
            query = quant_session.query(TradingDayList)
            trade_days = pd.read_sql(query.statement, query.session.bind)

        dts = trade_days[(trade_days.datetime >= datetime.date(2010, 1, 1)) &
                         (trade_days.datetime < pd.to_datetime(end_date, infer_datetime_format=True))].datetime.tolist()
        dts.reverse()
        dts = [dt for dt in dts if dt.weekday() == 4]
        for date in dts:
            fsg = FundScoreGroupV1Process(DerivedDataHelper(), data_cycle)
            if fsg.process(date, date, pd.to_datetime(date).date()):
                print(f'{date} failed')
                break
            print(f'{date} done')


if __name__ == '__main__':
    date = '20201009'
    fsg = FundScoreGroupV1Process(DerivedDataHelper(), 1)
    fsg.process(date, date, pd.to_datetime(date).date())
    # FundScoreGroupV1Process.update_history('20201010', 5)



import pandas as pd
import numpy as np
import platform
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle
from ....data.struct import TAAParam
CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']

class AssetPainter(object):

    @staticmethod
    def plot_asset_weights(asset_weight_dic:dict, saa:dict):
        asset_weight_df = pd.DataFrame(asset_weight_dic).T
        res = []
        for dic, s in zip(asset_weight_df.to_dict('records'), asset_weight_df.sum(axis = 1).values):
            res.append({ k: v/ s for k, v in dic.items()})
        df = pd.DataFrame(res)
        cols = df.columns.tolist()
        cols = [col for col in cols if df[col].sum() != 0]
        df = df[cols]
        df.index = asset_weight_df.index
        l = df.columns
        core_assets = ['hs300','csi500','gem','sp500rmb']
        asset_in_c = [i for i in l if i in core_assets]
        asset_not_in_c = sorted([i for i in l if i not in core_assets and i != 'cash']) 
        if 'national_debt' in asset_not_in_c:
            asset_not_in_c.remove('national_debt')
            asset_not_in_c = ['national_debt'] + asset_not_in_c
        l = asset_in_c + asset_not_in_c + ['cash']
        df = df[l]
        df.plot.area(figsize=(18,9),legend=False,fontsize = 17)
        plt.legend(loc='lower left',fontsize = 17)
        plt.title('asset weights history', fontsize=20)
        st = [k+':'+str(round(v,2)) for k, v in saa.items() if v > 0]
        st = ' '.join(st)
        plt.suptitle(st,y=0.87,fontsize=17)
        plt.grid()
        plt.show()

    @staticmethod
    def plot_asset_mdd_period(asset_mv:pd.DataFrame, saa_weight:dict, index_price: pd.DataFrame, asset_weight:dict):
        df = asset_mv
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd = round(1 - mdd_part1.min(),4)
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        l = [k for k, v in saa_weight.items() if v > 0]
        if 'cash' in l:
            l.remove('cash')
        df = index_price.copy()[l].loc[mdd_date1:mdd_date2].fillna(method='bfill')
        if df.empty:
            return 
        df = df / df.iloc[0]
        asset_weight = asset_weight[mdd_date1]
        for col in df:
            df[col] = 1 - (1 - df[col].values )* asset_weight.get(col, 0)
        d1 = mdd_date1.strftime('%Y-%m-%d')
        d2 = mdd_date2.strftime('%Y-%m-%d')
        df.plot.line(figsize=(18,9),legend=False,fontsize = 17)
        l = pl.legend(loc='lower left',fontsize = 17)
        s = pl.title('asset bt during mdd period price of holding assets', fontsize=20)
        plt.suptitle(f'mdd : {mdd}, from {d1} to {d2}',y=0.87,fontsize=17)    
        plt.grid()
        plt.show()

    @staticmethod
    def plot_asset_mdd_amount(asset_mv:pd.DataFrame, index_price: pd.DataFrame, position_history:dict):    
        index_ret = index_price / index_price.shift(1) - 1
        df = asset_mv
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd = round(1 - mdd_part1.min(),4)
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        volume_df = pd.DataFrame(position_history).T
        volume_df = volume_df.loc[mdd_date1: mdd_date2]
        if volume_df.empty:
            return 
        amount_df = volume_df.drop(['cash'], axis=1) * index_price.loc[mdd_date1: mdd_date2] * index_ret.loc[mdd_date1: mdd_date2]
        select_cols = []
        for col in amount_df:
            if amount_df[col].sum() != 0:
                select_cols.append(col)
        amount_df = amount_df[select_cols].cumsum()
        amount_df.plot.line(figsize=(18,9))
        l = pl.legend(loc='lower left',fontsize = 17)
        s = pl.title('asset bt during mdd period money loss', fontsize=20)
        plt.grid()
        plt.show()

    @staticmethod
    def plot_asset_taa_saa(saa_mv:pd.DataFrame, taa_mv:pd.DataFrame, index_id:str, index_pct:pd.DataFrame):
        taa_mv = taa_mv.rename(columns = {'mv':'taa'})
        saa_mv = saa_mv.rename(columns = {'mv':'saa'})
        begin_d = index_pct[[index_id]].index.values[0]
        df1 = saa_mv[['saa']].join(taa_mv['taa'])
        df1 = (df1 / df1.iloc[0])
        df1.plot.line(figsize=(20,12),legend=False,fontsize = 17)
        l = pl.legend(loc='upper left',fontsize = 25)
        s = pl.title(f'{index_id} taa', fontsize=20)
        plt.grid()
        plt.show()

    @staticmethod
    def plot_taa_analysis(saa_mv:pd.DataFrame, taa_mv:pd.DataFrame, index_id:str, index_pct:pd.DataFrame, taa_params:TAAParam, index_price:pd.DataFrame):
        def _plot_pct_high(ax, high_i, df, price_col):
            start = mdates.date2num(high_i['start_date'])
            end = mdates.date2num(high_i['end_date'])
            width = end - start
            price_max = df.loc[high_i['start_date']: high_i['end_date']][price_col].max() 
            bottom = high_i['end_pct'] 
            height = price_max - bottom
            rect = Rectangle((start, bottom), width, height, color='greenyellow')
            ax.add_patch(rect)   

        def _plot_pct_low(ax, low_i, df, price_col):
            start = mdates.date2num(low_i['start_date'])
            end = mdates.date2num(low_i['end_date'])
            width = end - start
            price_max = df.loc[low_i['start_date']: low_i['end_date']][price_col].max() 
            bottom = low_i['start_pct'] 
            height = price_max - bottom
            rect = Rectangle((start, bottom), width, height, color='cornsilk')
            ax.add_patch(rect)   

        def _plot_dif(ax, low_i, df, is_add):
            start = mdates.date2num(low_i['start_date'])
            end = mdates.date2num(low_i['end_date'])
            width = end - start
            price_max = df.loc[low_i['start_date']: low_i['end_date']]['taa/saa'].max() 
            bottom = df.loc[low_i['start_date']: low_i['end_date']]['taa/saa'].min() 
            height = price_max - bottom
            if is_add:
                c = 'cornsilk'
            else:
                c = 'greenyellow'
            rect = Rectangle((start, bottom), width, height, color=c)
            ax.add_patch(rect)   

        
        high_threshold = taa_params.HighThreshold
        low_threshold = taa_params.LowThreshold
        high_stop = taa_params.HighStop
        low_stop = taa_params.LowStop
        high_minus = taa_params.HighMinus
        low_plus = taa_params.LowPlus

        pct_col = index_id + '_pct'
        price_col = index_id + '_price'
        pct_df = index_pct[[index_id]].rename(columns= {index_id:f'{index_id}_pct'})
        price_df = index_price[['hs300','csi500','gem','sp500rmb']].rename(columns= {   
                                                                            'hs300':'hs300_price',
                                                                            'csi500':'csi500_price',
                                                                            'gem':'gem_price',
                                                                            'sp500rmb':'sp500rmb_price'
                                                                            })
        df = pct_df[[pct_col]].join(price_df[[price_col]]).dropna()
        df[price_col] = df[price_col] / df[price_col][0]

        # taa high part
        taa_con = False
        high_minus_periods = []
        for d, r in df.iterrows():
            if taa_con == False and r[pct_col] >= high_threshold:
                taa_con = True
                dic = {'start_date': d}
                dic['start_price'] = r[price_col]
                dic['start_pct'] = r[pct_col]
            elif taa_con == True and r[pct_col] <= high_stop:
                dic['end_date'] = d
                dic['end_price'] = r[price_col]
                dic['end_pct'] = r[pct_col]
                dic['high_avoid_pct'] = dic['end_price'] / dic['start_price'] - 1
                high_minus_periods.append(dic)
                taa_con = False
            if d == df.index[-1] and taa_con == True:
                dic['end_date'] = d
                dic['end_price'] = r[price_col]
                dic['end_pct'] = r[pct_col]
                dic['high_avoid_pct'] = dic['end_price'] / dic['start_price'] - 1
                high_minus_periods.append(dic)

        # taa low part
        taa_con = False
        low_plus_periods = []
        for d, r in df.iterrows():
            if taa_con == False and r[pct_col] <= low_threshold:
                taa_con = True
                dic = {'start_date': d}
                dic['start_price'] = r[price_col]
                dic['start_pct'] = r[pct_col]
            elif taa_con == True and r[pct_col] >= low_stop:
                dic['end_date'] = d
                dic['end_price'] = r[price_col]
                dic['end_pct'] = r[pct_col]
                dic['low_gain_pct'] = dic['end_price'] / dic['start_price'] - 1
                low_plus_periods.append(dic)
                taa_con = False
            if d == df.index[-1] and taa_con == True:
                dic['end_date'] = d
                dic['end_price'] = r[price_col]
                dic['end_pct'] = r[pct_col]
                dic['low_gain_pct'] = dic['end_price'] / dic['start_price'] - 1
                low_plus_periods.append(dic)

        fig, ax = plt.subplots(figsize= [20,12])
        plt.plot(df.index, df[pct_col], label = pct_col, linewidth=2.0)
        plt.plot(df.index, df[price_col], label = price_col, linewidth=2.0)
        plt.legend(fontsize=20 ,loc = 'upper left')
        plt.title(f'{index_id} TAA ', fontsize=30)
        plt.suptitle(f'high_threshold {high_threshold}  low_threshold {low_threshold}  high_stop {high_stop} low_stop {low_stop}',  
                    y=0.87, fontsize=18)
        for high_i in high_minus_periods:
            _plot_pct_high(ax, high_i, df, price_col)
        for low_i in low_plus_periods:
            _plot_pct_low(ax, low_i, df, price_col)
        plt.grid()
        plt.show()

        if taa_mv is not None and saa_mv is not None:
            taa_mv = taa_mv.rename(columns = {'mv':'taa'})
            saa_mv = saa_mv.rename(columns = {'mv':'saa'})
            df = saa_mv[['saa']].join(taa_mv['taa'])
            df = (df / df.iloc[0])
            df['taa/saa'] = df['taa'] / df['saa']
            fig, ax = plt.subplots(figsize= [20,12])
            plt.plot(df.index, df['taa/saa'], label = 'taa/saa', linewidth=2.0)
            plt.legend(fontsize=20 ,loc = 'upper left')
            plt.title(f'taa / ssa mv ', fontsize=30)
            for high_i in high_minus_periods:
                _plot_dif(ax, high_i, df, False)
            for low_i in low_plus_periods:
                _plot_dif(ax, low_i, df, True)
            pl.grid()
            plt.show()

    @staticmethod
    def plot_saa_rebalance_on_mv(asset_mv:pd.DataFrame,
                                rebalance_date:list,
                                MinActionAmtDiff:float,
                                annual_ret:float,
                                mdd:float,
                                ret_over_mdd:float):
        p = asset_mv['mv'].plot.line(figsize=(12,9), color = 'dodgerblue')
        m = MinActionAmtDiff
        annual_ret = round(annual_ret,4)
        mdd = round(mdd,4)
        ret_over_mdd=round(ret_over_mdd,4)
        t = plt.title(f'market value with AmtDiff {m} annual_ret {annual_ret} mdd {mdd} ret_over_mdd {ret_over_mdd} ', fontsize=20)
        for d, r in asset_mv.loc[rebalance_date,:].iterrows():
            pl.scatter(x=d, y=r.mv, c='r',s=50)
        pl.grid()
        pl.show()

    @staticmethod
    def plot_index_price_ratio(df,t):
        df.index = [str(i) for i in df.index]
        dic = ((df.iloc[-1]-1) * 10000).astype(int).to_dict()
        m = int(np.mean(list(dic.values())))
        _st = sorted(dic.items(), key=lambda x:x[1], reverse=True)
        st_1 = [f'mean diff: {m} BP  ']
        st_2 = []
        for i in _st:
            if i[1] > 0:
                st_1.append(f'{i[0]}:{i[1]} ')
            else:    
                st_2.append(f'{i[0]}:{i[1]} ')
        st = ''.join(st_1)+'\n'+''.join(st_2)
        df.plot.line(figsize=(14,8),fontsize=15)
        plt.legend(fontsize=13, loc='lower left')
        plt.title(t,fontsize=22)
        plt.suptitle(st, y=0.87, fontsize=18)
        plt.grid()
        plt.show()

        


import datetime
import pandas as pd
from . import DataManager
from .wrapper_stock import StockDailyWrapper
from ...structure import TickerInfo

class StockDataManager(DataManager):

    def __init__(self,data_path=None,start_time=None,end_time=None,universe=None,datatype=None,stock_pool=None,price_type=None,factor_path=None,factor_name=None,vwap_limit=None):
        self.wrapper = datatype(data_path,price_type,stock_pool,start_time,end_time,factor_path,factor_name,vwap_limit) if datatype == StockDailyWrapper else datatype(data_path)
        self.universe = universe
        self.frames = None
        self.start_time = start_time if isinstance(start_time, datetime.datetime) else datetime.datetime.strptime(start_time,'%Y%m%d')
        self.end_time = end_time if isinstance(end_time, datetime.datetime) else datetime.datetime.strptime(end_time,'%Y%m%d')
        self.t_info = 0 if datatype == StockDailyWrapper else self.init_t_info() 
        self.modify_df()
        self.stock_pool=stock_pool
        
    def init(self):
        pass

    def get_data(self):
        return self._get_frames(self.start_time, self.end_time, self.universe)

    def parse_row(self, index, row):
        return self.wrapper.parse(index, row)

    def find_start_file(self, ticker):
        start_day_str = self.start_time.strftime('%Y%m%d')
        for i in self.wrapper.meta_info[ticker]['data_list'][:-1]:
            next_file_index = self.wrapper.meta_info[ticker]['data_list'].index(i)+1
            if start_day_str >= i and start_day_str < self.wrapper.meta_info[ticker]['data_list'][next_file_index]:
                return i
        assert False, 'data not exist'

    def _get_months(self, start_dt, end_dt):
        months = set([])
        dt = start_dt
        while dt <= end_dt:
            months.add(dt.strftime('%Y%m'))
            dt += datetime.timedelta(days=1)
        months = sorted(months)
        return months

    def init_t_info(self):
        dic = {}
        if not self.universe:
            return None
        for ticker in self.universe:
            dic[ticker] = TickerInfo()
            dic[ticker].loc = 0
            dic[ticker].file = self.find_start_file(ticker)
            dic[ticker].file_reader_time = self.start_time
            dic[ticker].data_list = self.wrapper.meta_info[ticker]['data_list']
            dic[ticker].data_bite = int(self.wrapper.meta_info[ticker]['data_bite'])
        return dic

    def modify_df(self):
        if not self.universe:
            return None

        for ticker in self.universe:
            self.t_info[ticker].df   = self.load_df_bite(ticker)
            if self.t_info[ticker].df.empty:
                self.t_info[ticker].len  = 0
                self.t_info[ticker].file_end_time  = 0
            else:
                self.t_info[ticker].len  = self.t_info[ticker].df.shape[0]
                self.t_info[ticker].file_end_time  = self.t_info[ticker].df.index[-1]

    def get_current_time(self):
        return  [self.t_info[i].df.index[self.t_info[i].loc] for i in self.universe]

    def load_df_bite(self, ticker):
        df_item = self.wrapper.load_data(self.t_info[ticker].file, ticker)
        df_item['ticker'] = ticker
        df_item.sort_index(inplace=True)
        if df_item.empty:
            self.t_info[ticker].file_end_time  = 0
        else:
            self.t_info[ticker].file_end_time = df_item.index[-1]
        begin_time = max(self.t_info[ticker].file_reader_time, df_item.index[0])
        finish_time = begin_time + pd.offsets.Day(self.t_info[ticker].data_bite)
        self.t_info[ticker].file_reader_time = finish_time
        if self.t_info[ticker].file_reader_time > self.t_info[ticker].file_end_time:
            self.t_info[ticker].file_reader_time = self.start_time
            idx = self.t_info[ticker].data_list.index(self.t_info[ticker].file) + 1
            self.t_info[ticker].file = self.t_info[ticker].data_list[idx]    
        begin_time  = begin_time.strftime('%Y-%m-%d') + ' 00:00:00'
        finish_time = finish_time + pd.offsets.Day(1)   
        finish_time = finish_time.strftime('%Y-%m-%d') + ' 00:00:00'
        df_item = df_item[df_item.index >= begin_time]
        df_item = df_item[df_item.index <= finish_time]    
        return df_item

    def reload(self):
        for ticker in self.universe:
            if self.t_info[ticker].loc == self.t_info[ticker].len:
                df_item = pd.DataFrame()
                while df_item.empty:
                    df_item = self.load_df_bite(ticker)
                    if not df_item.empty:
                        self.t_info[ticker].df = df_item
                        self.t_info[ticker].loc = 0
                        self.t_info[ticker].len = df_item.shape[0]
                        self.t_info[ticker].file_end_time = df_item.index[-1]
                    elif df_item.empty:
                        if self.t_info[ticker].file_reader_time >  self.t_info[ticker].file_end_time:
                            idx = self.t_info[ticker].data_list.index(self.t_info[ticker].file) + 1
                            try:
                                self.t_info[ticker].file = self.t_info[ticker].data_list[idx]               
                                self.t_info[ticker].file_reader_time = self.start_time
                            except:
                                print('download more data')
                                break
                        df_item = self.load_df_bite(ticker)

    def get_next_tick(self):
        self.reload()
        # if time > end_time,  stop
        time_list = self.get_current_time()
        current_time = min(time_list)
        if current_time > self.end_time: 
            return
        # normal send and renew loc_list
        i = time_list.index(current_time)
        ticker = self.universe[i]
        row = self.t_info[ticker].df.iloc[self.t_info[ticker].loc]
        self.t_info[ticker].loc += 1
        self.current_index = row.name
        return self.parse_row(row.name, row)
        
    def get_next_day(self):
        # remove st and suspended from pool_today 
        # build today factor filter by pool_today
        try:        
            date = self.wrapper.date_list[self.t_info]
            self.stock_pool_today = self.wrapper.stock_pool_data[date]
            self.st_today = self.wrapper.st_data[date]
            self.suspended_today = self.wrapper.suspended_data[date]
            ban_list = list(set(self.st_today + self.suspended_today))
            self.stock_pool_today = [item for item in self.stock_pool_today if item not in ban_list]
            self.factor_today = self.wrapper.factor_data[date]
            self.t_info += 1      
            self.factor_today = {k: v for k, v in self.factor_today.items() if k in self.stock_pool_today} 
            self.rm_l_today = self.wrapper.remove_limit_vwap_stock(date)
            return self.wrapper.price_data[date]    
        except:
            return None

    def _get_frames(self, start_time=None, end_time=None, universe=None):
        months = self._get_months(start_time, end_time)
        dfs = []
        for ticker in universe:
            for month in months:
                df_item = self.wrapper.load_data(month, ticker)
                df_item['ticker'] = ticker
                dfs.append(df_item)
        dfs = pd.concat(dfs)
        dfs.sort_index(inplace=True)
        dfs = dfs[dfs.index >= start_time.strftime('%Y-%m-%d') + ' 00:00:00']
        dfs = dfs[dfs.index <= end_time.strftime('%Y-%m-%d') + ' 23:59:59']
        return dfs

    def terminate(self):
        pass


    
    